 /*
  * UAE - The Un*x Amiga Emulator
  *
  * joystick/mouse emulation
  *
  * Copyright 2001, 2002 Toni Wilen
  *
  * new fetures:
  * - very configurable (and very complex to configure :)
  * - supports multiple native input devices (joysticks and mice)
  * - supports mapping joystick/mouse buttons to keys and vice versa
  * - joystick mouse emulation (supports both ports)
  * - full cd32 pad support (supports both ports)
  * - fully backward compatible with old joystick/mouse configuration
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "inputdevice.h"
#include "keybuf.h"
#include "custom.h"
#include "xwin.h"
#include "drawing.h"
#include "memory.h"
#include "events.h"
#include "newcpu.h"

/* seems to be good idea to allow following being configurable */
static int joymouse_multiplier = 20;
static int joymouse_deadzone = 33;
static int joymouse_speed = 10;
static int autofire_framecnt = 10;

#define DIR_LEFT 1
#define DIR_RIGHT 2
#define DIR_UP 4
#define DIR_DOWN 8

struct inputevent {
    char *name;
    int allow_mask;
    int unit;
    int data;
};

#define DEFEVENT(A, B, C, D, E) { B, C, D, E },
struct inputevent events[] = {
{0, 0, 0, 0},
#include "inputevents.def"
{0, 0, 0, 0}
};
#undef DEFEVENT

struct uae_input_device {
    int nr;
    int buttons_eventid[MAX_INPUT_DEVICE_BUTTONS];
    int buttons_flags[MAX_INPUT_DEVICE_BUTTONS];
    uae_u32 buttonmask;
    int axles_eventid[MAX_INPUT_DEVICE_BUTTONS];
    int axles_flags[MAX_INPUT_DEVICE_BUTTONS];
    int states[MAX_INPUT_DEVICE_BUTTONS];
};

static int isdevice (struct uae_input_device *id)
{
    int i;
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if (id->buttons_eventid[i] > 0) return 1;
    }
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if (id->axles_eventid[i] > 0) return 1;
    }
    return 0;
}

static struct uae_input_device joysticks[MAX_INPUT_DEVICES];
static struct uae_input_device mice[MAX_INPUT_DEVICES];
static uae_u8 keyboard_mapping[256];

static int mouse_x[2], mouse_y[2];
static int mouse_x_delta[2], mouse_y_delta[2];
static int joybutton[2];
static unsigned int joydir[2];
static int cd32_shifter[2];
static int cd32_pad_enabled[2];

static int oldmx[2], oldmy[2];
static int oleft[2], oright[2], otop[2], obot[2];

static int max_joysticks, max_mice;

#define INPUT_QUEUE_SIZE 16
struct input_queue_struct {
    int event, storedstate, state, max, framecnt, nextframecnt;
};
static struct input_queue_struct input_queue[INPUT_QUEUE_SIZE];

static void out_config (FILE *f, int num, char *s1, char *s2)
{
//    fprintf (f, "%s%s\n", s1, s2);
    write_log ("%d=%s%s\n", num, s1,s2);
}

static void write_config (FILE *f, char *name, struct uae_input_device *id)
{
    char tmp1[100],tmp2[100];
    int i, event;

    if (!isdevice (id))
	return;
    sprintf (tmp1, "%s.%d.axle.", name, id->nr);
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if ((event = id->axles_eventid[i]) > 0) {
	    sprintf (tmp2, "%s.%d", events[event].name, id->axles_flags[i]);
	    out_config (f, i, tmp1, tmp2);
	}
    }
    sprintf (tmp1, "%s.%d.button." ,name, id->nr);
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if ((event = id->buttons_eventid[i]) > 0) {
	    sprintf (tmp2, "%s.%d", events[event].name, id->buttons_flags[i]);
	    out_config (f, i, tmp1, tmp2);
	}
    }
}

void write_inputdevice_config (FILE *f)
{
    int i;
    for (i = 0; i < MAX_INPUT_DEVICES; i++)
	write_config (f, "joystick", &joysticks[i]);
    for (i = 0; i < MAX_INPUT_DEVICES; i++)
	write_config (f, "mouse", &mice[i]);
}

static int getnum (char **pp)
{
    char *p = *pp;
    int v = atol (p);
    while (*p != 0 && *p !='.') p++;
    *pp = p;
    return v;
}
static char *getstring (char **pp)
{
    int i;
    static char str[100];
    char *p = *pp;

    i = 0;
    while (*p != 0 && *p !='.') str[i++] = *p++;
    *pp = p;
    return str;
}

void read_inputdevice_config (char *option, char *value)
{
/*
    struct uae_input_device *id = 0;
    int devnum, num, button, flags, i;
    char *p, *p2;

    if (memcmp (option, "mouse.", 6) == 0) {
	id = mice;
	p = option + 6;
    } else if (memcmp (option, "joystick.", 9) == 0) {
	id = joysticks;
	p = option + 9;
    }
    if (!id)
	return;
    devnum = getnum (&p);
    if (devnum < 0 || devnum >= MAX_INPUT_DEVICES)
	return;
    p2 = getstring (&p);
    if (!p2)
	return;
    button = -1;
    if (!strcmp (p2, "axle"))
	button = 0;
    else if(!strcmp (p2, "button"))
	button = 1;
    if (button < 0)
	return;
    num = getnum (&p);
    if (devnum < 0 || devnum >= MAX_INPUT_DEVICES)
	return;
    p2 = getstring (&p);
    flags = getnum (&p);
    i = 1;
    while (events[i].name) {
	if (!strcmp (events[i].name, p2))
	    break;
	i++;
    }
    if (!events[i].name)
	return;
*/
}

/* Mousehack stuff */

#define defstepx (1<<16)
#define defstepy (1<<16)
#define defxoffs 0
#define defyoffs 0

static const int docal = 60, xcaloff = 40, ycaloff = 20;
static const int calweight = 3;
static int lastsampledmx, lastsampledmy;
static int lastspr0x,lastspr0y,lastdiffx,lastdiffy,spr0pos,spr0ctl;
static int mstepx,mstepy,xoffs=defxoffs,yoffs=defyoffs;
static int sprvbfl;

static int lastmx, lastmy;
static int newmousecounters;
static enum mousestate mousestate;

void mousehack_handle (int sprctl, int sprpos)
{		    
    if (!sprvbfl && ((sprpos & 0xff) << 2) > 2 * DISPLAY_LEFT_SHIFT) {
	spr0ctl = sprctl;
	spr0pos = sprpos;
	sprvbfl = 2;
    }
}

static void mousehack_setunknown (void)
{
    mousestate = mousehack_unknown;
}

static void mousehack_setdontcare (void)
{
    if (mousestate == mousehack_dontcare)
	return;

    write_log ("Don't care mouse mode set\n");
    mousestate = mousehack_dontcare;
    lastspr0x = lastmx; lastspr0y = lastmy;
    mstepx = defstepx; mstepy = defstepy;
}

static void mousehack_setfollow (void)
{
    if (mousestate == mousehack_follow)
	return;

    write_log ("Follow sprite mode set\n");
    mousestate = mousehack_follow;
    lastdiffx = lastdiffy = 0;
    sprvbfl = 0;
    spr0ctl = spr0pos = 0;
    mstepx = defstepx; mstepy = defstepy;
}

void mousehack_set (int state)
{
    switch (state)
    {
	case mousehack_dontcare:
	mousehack_setdontcare();
	break;
	case mousehack_follow:
	mousehack_setfollow();
	break;
	default:
	mousestate = state;
	break;
    }
}

uae_u32 mousehack_helper (void)
{
    int mousexpos, mouseypos;

#ifdef PICASSO96
    if (picasso_on) {
	mousexpos = lastmx - picasso96_state.XOffset;
	mouseypos = lastmy - picasso96_state.YOffset;
    } else
#endif
    {
	if (lastmy >= gfxvidinfo.height)
	    lastmy = gfxvidinfo.height - 1;
	mouseypos = coord_native_to_amiga_y (lastmy) << 1;
	mousexpos = coord_native_to_amiga_x (lastmx);
    }

    switch (m68k_dreg (regs, 0)) {
    case 0:
	return ievent_alive ? -1 : needmousehack ();
    case 1:
	ievent_alive = 10;
	return mousexpos;
    case 2:
	return mouseypos;
    }
    return 0;
}

void togglemouse (void)
{
    switch (mousestate) {
     case mousehack_dontcare: mousehack_setfollow (); break;
     case mousehack_follow: mousehack_setdontcare (); break;
     default: break; /* Nnnnnghh! */
    }
}

STATIC_INLINE int adjust (int val)
{
    if (val > 127)
	return 127;
    else if (val < -127)
	return -127;
    return val;
}

static void do_mouse_hack (void)
{
    int spr0x = ((spr0pos & 0xff) << 2) | ((spr0ctl & 1) << 1);
    int spr0y = ((spr0pos >> 8) | ((spr0ctl & 4) << 6)) << 1;
    int diffx, diffy;

    if (ievent_alive > 0) {
	mouse_x[0] = mouse_y[0] = 0;
	return;
    }
    switch (mousestate) {
    case mousehack_normal:
	diffx = lastmx - lastsampledmx;
	diffy = lastmy - lastsampledmy;
	if (!newmousecounters) {
	    if (diffx > 127) diffx = 127;
	    if (diffx < -127) diffx = -127;
	    mouse_x[0] += diffx;
	    if (diffy > 127) diffy = 127;
	    if (diffy < -127) diffy = -127;
	    mouse_y[0] += diffy;
	}
	lastsampledmx += diffx; lastsampledmy += diffy;
	break;

    case mousehack_dontcare:
	diffx = adjust (((lastmx - lastspr0x) * mstepx) >> 16);
	diffy = adjust (((lastmy - lastspr0y) * mstepy) >> 16);
	lastspr0x = lastmx; lastspr0y = lastmy;
	mouse_x[0] += diffx; mouse_y[0] += diffy;
	break;

    case mousehack_follow:
	if (sprvbfl && sprvbfl-- > 1) {
	    int mousexpos, mouseypos;

	    if ((lastdiffx > docal || lastdiffx < -docal)
		&& lastspr0x != spr0x
		&& spr0x > plfstrt*4 + 34 + xcaloff
		&& spr0x < plfstop*4 - xcaloff)
	    {
		int val = (lastdiffx << 16) / (spr0x - lastspr0x);
		if (val >= 0x8000)
		    mstepx = (mstepx * (calweight - 1) + val) / calweight;
	    }
	    if ((lastdiffy > docal || lastdiffy < -docal)
		&& lastspr0y != spr0y
		&& spr0y > plffirstline + ycaloff
		&& spr0y < plflastline - ycaloff)
	    {
		int val = (lastdiffy << 16) / (spr0y - lastspr0y);
		if (val >= 0x8000)
		    mstepy = (mstepy * (calweight - 1) + val) / calweight;
	    }
	    if (lastmy >= gfxvidinfo.height)
		lastmy = gfxvidinfo.height-1;
	    mouseypos = coord_native_to_amiga_y (lastmy) << 1;
	    mousexpos = coord_native_to_amiga_x (lastmx);
	    diffx = adjust ((((mousexpos + xoffs - spr0x) & ~1) * mstepx) >> 16);
	    diffy = adjust ((((mouseypos + yoffs - spr0y) & ~1) * mstepy) >> 16);
	    lastspr0x = spr0x; lastspr0y = spr0y;
	    lastdiffx = diffx; lastdiffy = diffy;
	    mouse_x[0] += diffx; mouse_y[0] += diffy;
	}
	break;
	
    default:
	abort ();
    }
}

int getbuttonstate (int joy, int button)
{
    return joybutton[joy] & (1 << button);
}

int getjoystate (int joy)
{
    if (JSEM_ISMOUSE (joy, &currprefs)) {
	if (joy == 0) do_mouse_hack ();
	mouse_x[joy] += mouse_x_delta[joy];
	mouse_y[joy] += mouse_y_delta[joy];
	return ((uae_u8)mouse_x[joy]) + ((uae_u16)mouse_y[joy] << 8);
    }
    return joydir[joy];
}
uae_u16 JOY0DAT (void)
{
    return getjoystate (0);
}
uae_u16 JOY1DAT (void)
{
    return getjoystate (1);
}

void JOYTEST (uae_u16 v)
{
    mouse_x[0] = mouse_x[1] = v & 0xFC;
    mouse_y[0] = mouse_y[1] = (v >> 8) & 0xFC;
}

static uae_u16 handle_joystick_potdat (int joy)
{
    static uae_u16 cnt[2] = { 0, 0 };
    if (getbuttonstate (joy, JOYBUTTON_2))
        cnt[joy] = ((cnt[joy] + 1) & 0xFF) | (cnt[joy] & 0xFF00);
    if (getbuttonstate (joy, JOYBUTTON_3))
        cnt[joy] += 0x100;
    return cnt[joy];
}

uae_u16 POT0DAT (void)
{
    return handle_joystick_potdat (0);
}
uae_u16 POT1DAT (void)
{
    return handle_joystick_potdat (1);
}

uae_u8 handle_joystick_buttons (uae_u8 dra)
{
    uae_u8 but = 0;
    if (!getbuttonstate (0, JOYBUTTON_1) && !getbuttonstate (0, JOYBUTTON_CD32_RED))
        but |= 0x40;
    if (!getbuttonstate (1, JOYBUTTON_1) && !getbuttonstate (1, JOYBUTTON_CD32_RED))
        but |= 0x80;
    return but;
}

/* joystick 1 button 1 is used as a output for incrementing shift register */
void handle_cd32_joystick_cia (uae_u8 pra, uae_u8 dra)
{
    static int oldstate[2];
    if ((dra & 0x80) && (pra & 0x80) != oldstate[1]) {
	if (!(pra & 0x80))
	    cd32_shifter[1]--;
	oldstate[1] = pra & 0x80;
    }
    if ((dra & 0x40) && (pra & 0x40) != oldstate[0]) {
	if (!(pra & 0x40))
	    cd32_shifter[0]--;
	oldstate[0] = pra & 0x40;
    }
}

/* joystick port 1 button 3 is used as a output for clearing shift register */
static void handle_cd32_joystick_potgo (uae_u16 potgo)
{
    static int oldstate[2];
    if ((potgo & 0x3000) == 0x2000) {
	if (!oldstate[1]) {
	    cd32_shifter[1] = 8;
	    oldstate[1] = 1;
	}
    } else
	oldstate[1] = 0;
    if ((potgo & 0x0300) == 0x0200) {
	if (!oldstate[0]) {
	    cd32_shifter[0] = 8;
	    oldstate[0] = 1;
	}
    } else
	oldstate[0] = 0;
}

/* joystick port 1 button 2 is input for button state */
static uae_u16 handle_joystick_potgor (uae_u16 potgor)
{
    if (cd32_pad_enabled[1] && !(potgo_value & 0x1000)) {
	if (cd32_shifter[1] == 0)
	    potgor &= ~0x4000;
	else if (cd32_shifter[1] >= 2 && (joybutton[1] & ((1 << JOYBUTTON_CD32_PLAY) << (cd32_shifter[1] - 2))))
	    potgor &= ~0x4000;
    } else {
        if (getbuttonstate (1, JOYBUTTON_2))
	    potgor &= ~0x4000;
	if (getbuttonstate (1, JOYBUTTON_3))
	    potgor &= ~0x1000;
    }
    if (cd32_pad_enabled[0] && !(potgo_value & 0x0100)) {
        if (cd32_shifter[0] == 0)
	    potgor &= ~0x0400;
	else if (cd32_shifter[0] >=2 && (joybutton[0] & ((1 << JOYBUTTON_CD32_PLAY) << (cd32_shifter[0] - 2))))
	    potgor &= ~0x0400;
    } else {
        if (getbuttonstate (0, JOYBUTTON_2))
	    potgor &= ~0x0400;
	if (getbuttonstate (0, JOYBUTTON_3))
	    potgor &= ~0x0100;
    }
    return potgor;
}

uae_u16 potgo_value;
void POTGO (uae_u16 v)
{
    potgo_value = v;
    handle_cd32_joystick_potgo (v);
}

uae_u16 POTGOR (void)
{
    uae_u16 v = (potgo_value | (potgo_value >> 1)) & 0x5500;
    v |= (~potgo_value & 0xAA00) >> 1;
    return handle_joystick_potgor (v);
}

static int check_input_queue (int event)
{
    struct input_queue_struct *iq;
    int i;
    for (i = 0; i < INPUT_QUEUE_SIZE; i++) {
        iq = &input_queue[i];
        if (iq->event == event) return i;
    }
    return -1;
}

static void handle_input_event (int nr, int state, int max, int autorepeat);

static void queue_input_event (int event, int state, int max, int framecnt, int autofire)
{
    struct input_queue_struct *iq;
    int i = check_input_queue (event);

    if (state < 0 && i >= 0) {
        iq = &input_queue[i];
	iq->nextframecnt = -1;
	iq->framecnt = -1;
	iq->event = 0;
	if (iq->state == 0)
	    handle_input_event (event, 0, 1, 0);
    } else if (i < 0) {
	for (i = 0; i < INPUT_QUEUE_SIZE; i++) {
	    iq = &input_queue[i];
	    if (iq->framecnt < 0) break;
	}
	if (i == INPUT_QUEUE_SIZE) {
	    write_log ("input queue overflow\n");
	    return;
	}
	iq->event = event;
	iq->state = iq->storedstate = state;
	iq->max = max;
	iq->framecnt = framecnt;
	iq->nextframecnt = autofire > 0 ? framecnt : -1;
    }
}

static void handle_input_event (int nr, int state, int max, int autorepeat)
{
    struct inputevent *ie;
    int joy;
    
    if (nr <= 0) return;
    ie = &events[nr];
    if (autorepeat) {
	if (state)
	    queue_input_event (nr, state, max, autofire_framecnt, 1);
	    else
	    queue_input_event (nr, -1, 0, 0, 1);
    }
    switch (ie->unit & 3)
    {
	case 1: /* ->JOY1 */
	case 2: /* ->JOY2 */
	    joy = (ie->unit & 3) - 1;
	    if (ie->unit & 4) {
		if (state)
		    joybutton[joy] |= 1 << ie->data;
		else
		    joybutton[joy] &= ~(1 << ie->data);
	    } else if (ie->unit & 8) {
		/* analog stick mouse emulation */
		int delta;
		int deadzone = joymouse_deadzone * max / 100;
		if (state < deadzone && state > -deadzone) {
		    state = 0;
		} else if (state < 0) {
		    state += deadzone;
		} else {
		    state -= deadzone;
		}
		max -= deadzone;
		delta = state * joymouse_multiplier / max;
		if (ie->data)
		    mouse_y_delta[joy] = delta;
		    else
		    mouse_x_delta[joy] = delta;
	    } else if (ie->unit & 32) {
		/* button mouse emulation */
		if (state && (ie->data & DIR_LEFT))
		    mouse_x_delta[joy] = -joymouse_speed;
		else if (state && (ie->data & DIR_RIGHT))
		    mouse_x_delta[joy] = joymouse_speed;
		else
		    mouse_x_delta[joy] = 0;
		if (state && (ie->data & DIR_UP))
		    mouse_y_delta[joy] = -joymouse_speed;
		else if (state && (ie->data & DIR_DOWN))
		    mouse_y_delta[joy] = joymouse_speed;
		else
		    mouse_y_delta[joy] = 0;
	    } else {
	        int left = oleft[joy], right = oright[joy], top = otop[joy], bot = obot[joy];
		if (ie->unit & 16) {
		    /* button to axle mapping */
		    if (ie->data & DIR_LEFT) left = oleft[joy] = state ? 1 : 0;
		    if (ie->data & DIR_RIGHT) right = oright[joy] = state ? 1 : 0;
		    if (ie->data & DIR_UP) top = otop[joy] = state ? 1 : 0;
		    if (ie->data & DIR_DOWN) bot = obot[joy] = state ? 1 : 0;
		} else {
    		    /* "normal" joystick axle */
		    int neg = state < -max / 2  ? 1 : 0;
		    int pos = state > max / 2 ? 1 : 0;
		    if (ie->data & DIR_LEFT) left = oleft[joy] = neg;
		    if (ie->data & DIR_RIGHT) right = oright[joy] = pos;
		    if (ie->data & DIR_UP) top = otop[joy] = neg;
		    if (ie->data & DIR_DOWN) bot = obot[joy] = pos;
		}
	        if (left)
		    top = !top;
		if (right)
		    bot = !bot;
		joydir[joy] = bot | (right << 1) | (top << 8) | (left << 9);
	    }
	break;
	case 0: /* ->KEY */
	    record_key ((ie->data << 1) | (state ? 0x00 : 0x01));
	break;
    }
}

void inputdevice_vsync (void)
{
    struct input_queue_struct *iq;
    int i;

    for (i = 0; i < INPUT_QUEUE_SIZE; i++) {
	iq = &input_queue[i];
	if (iq->framecnt > 0) {
	    iq->framecnt--;
	    if (iq->framecnt == 0) {
		handle_input_event (iq->event, iq->state, iq->max, 0);
		iq->framecnt = iq->nextframecnt;
		if (iq->state) iq->state = 0; else iq->state = iq->storedstate;
	    }
	}
    }
    read_joysticks ();
    read_mice ();
}


STATIC_INLINE void setbuttonstateall (struct uae_input_device *id, int button, int state)
{
    int event = id->buttons_eventid[button];
    int autorepeat = (id->buttons_flags[button] & ID_FLAG_AUTOREPEAT) ? 1 : 0;
    if (button >= MAX_INPUT_DEVICE_BUTTONS)
	return;
    if (state < 0) {
	handle_input_event (event, 1, 1, 0);
	queue_input_event (event, 0, 1, 1, 0); /* send release event next frame */
    } else {
        uae_u32 mask = 1 << button;
	uae_u32 omask = id->buttonmask & mask;
	uae_u32 nmask = (state ? 1 : 0) << button;
	if ((omask ^ nmask) & mask) {
	    handle_input_event (event, state, 1, autorepeat);
	    if (state)
		id->buttonmask |= mask;
	    else
		id->buttonmask &= ~mask;
	}
    }
}

/* Call this function when host machine's joystick/joypad/etc button state changes
 * This function translates button events to Amiga joybutton/joyaxle/keyboard events
 * TODO: configuration GUI
 */

/* button states:
 * state = -1 -> mouse wheel turned or similar (button without release)
 * state = 1 -> button pressed
 * state = 0 -> button released
 */

void setjoybuttonstate (int joy, int button, int state)
{
    setbuttonstateall (&joysticks[joy], button, state ? 1 : 0);
}

/* buttonmask = 1 = normal toggle button, 0 = mouse wheel turn or similar
 */
void setjoybuttonstateall (int joy, uae_u32 buttonbits, uae_u32 buttonmask)
{
    int i;
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if (buttonmask & (1 << i)) {
	    setbuttonstateall (&joysticks[joy], i, (buttonbits & (1 << i)) ? 1 : 0);
	} else if (buttonbits & (1 << i)) {
	    setbuttonstateall (&joysticks[joy], i, -1);
	}
    }
}
/* mouse buttons (just like joystick buttons)
 */
void setmousebuttonstateall (int mouse, uae_u32 buttonbits, uae_u32 buttonmask)
{
    int i;
    for (i = 0; i < MAX_INPUT_DEVICE_BUTTONS; i++) {
	if (buttonmask & (1 << i)) {
	    setbuttonstateall (&mice[mouse], i, (buttonbits & (1 << i)) ? 1 : 0);
	} else if (buttonbits & (1 << i)) {
	    setbuttonstateall (&mice[mouse], i, -1);
	}
    }
}

void setmousebuttonstate (int mouse, int button, int state)
{
    setbuttonstateall (&mice[mouse], button, state);
}

/* same for joystick axles (analog or digital)
 * (0 = center, -max = full left/top, max = full right/bottom)
 */
void setjoystickstate (int joy, int axle, int state, int max)
{
    struct uae_input_device *id = &joysticks[joy];
    if (id->states[axle] == state) return;
    handle_input_event (id->axles_eventid[axle], state, max, id->axles_flags[axle]);
    id->states[axle] = state;
}
/* mouse position, absolute or relative (must be absolute for mouse=0)
 */
void setmousestate (int mouse, int x, int y, int isabs)
{
    int dx = 0, dy = 0;
    struct inputevent *e = &events[mice[mouse].axles_eventid[0]];

    if (e == &events[0]) return;
    mouse = (e->unit & 3) - 1;

    if (!isabs) {
	dx = adjust (dx);
	dy = adjust (dy);
	mouse_x[mouse] += dx;
	mouse_y[mouse] += dy;
    } else if (mouse == 0) {
	if (oldmx[mouse] >= 0)
	    dx = oldmx[mouse] - dx;
	if (oldmy[mouse] >= 0)
	    dy = oldmy[mouse] - dy;
	oldmx[mouse] = x;
	oldmy[mouse] = y;
        lastmx = oldmx[mouse];
        lastmy = oldmy[mouse];
    } else {
	dx = x - oldmx[mouse];
	dy = y - oldmy[mouse];
	oldmx[mouse] = x;
	oldmy[mouse] = y;
	mouse_x[mouse] += dx;
	mouse_y[mouse] += dy;
    }
}

static struct native2amigakey *keytranstable;

/* - detect required number of joysticks and mice from configuration data
 * - detect if CD32 pad emulation is needed
 */
extern int cd32_enabled;
static void scanjoyevents(void)
{
    int i, j;
    struct inputevent *e;
    struct native2amigakey *na = keytranstable;

    max_joysticks = max_mice = -1;
    cd32_pad_enabled[0] = cd32_pad_enabled[1] = 0;
    for (i = 0; i < MAX_INPUT_DEVICES; i++) {
	for (j = 0; j < MAX_INPUT_DEVICE_BUTTONS; j++) {
	    e = &events[joysticks[i].buttons_eventid[j]];
	    if ((e->unit & 4) && e->data >= JOYBUTTON_CD32_PLAY) cd32_pad_enabled[(e->unit & 3) - 1] = 1;
	    e = &events[mice[i].buttons_eventid[j]];
	    if ((e->unit & 4) && e->data >= JOYBUTTON_CD32_PLAY) cd32_pad_enabled[(e->unit & 3) - 1] = 1;
	    if (joysticks[i].buttons_eventid[j] > 0 && max_joysticks < i) max_joysticks = i;
	    if (mice[i].buttons_eventid[j] > 0 && max_mice < i) max_mice = i;
	}
	for (j = 0; j < MAX_INPUT_DEVICE_BUTTONS; j++) {
	    if (joysticks[i].axles_eventid[j] > 0 && max_joysticks < i) max_joysticks = i;
	    if (mice[i].axles_eventid[j] > 0 && max_mice < i) max_mice = i;
	}
    }
    while (na && na->event != 0) {
	if (na->event >= INPUTEVENT_JOY1_CD32_FIRST && na->event <= INPUTEVENT_JOY1_CD32_LAST)
	    cd32_pad_enabled[0] = 1;
	if (na->event >= INPUTEVENT_JOY2_CD32_FIRST && na->event <= INPUTEVENT_JOY2_CD32_LAST)
	    cd32_pad_enabled[1] = 1;
	na++;
    }
    if (!cd32_enabled) {
	cd32_pad_enabled[0] = 0;
	cd32_pad_enabled[1] = 0;
    }

    max_joysticks++;
    if (max_joysticks > nr_joysticks)
	max_joysticks = nr_joysticks;
    max_mice++;
    if (max_mice > nr_mice)
	max_mice = nr_mice;
}

static int geid (char *name)
{
    int i = 0;
    while (events[i].name) {
	if (!strcmp (events[i].name, name)) return i;
	i++;
    }
    write_log ("unknown input event '%s'\n", name);
    return 0;
}

void clearinputstate (void)
{
    int joy, i;

    joybutton[0] = joybutton[1] = 0;
    joydir[0] = joydir[1] = 0;
    oldmx[0] = oldmx[1] = -1;
    oldmy[0] = oldmy[1] = -1;
    cd32_shifter[0] = cd32_shifter[1] = 8;
    oleft[0] = oleft[1] = 0;
    oright[0] = oright[1] = 0;
    otop[0] = otop[1] = 0;
    obot[0] = obot[1] = 0;

    for (i = 0; i < INPUT_QUEUE_SIZE; i++)
	input_queue[i].framecnt = input_queue[i].nextframecnt = -1;

    memset (keyboard_mapping, 0, sizeof (keyboard_mapping));

    /* mouse compatibility code */
    if (JSEM_ISMOUSE (0, &currprefs)) {
	mice[0].axles_eventid[0] = INPUTEVENT_MOUSE1;
        mice[0].buttons_eventid[0] = INPUTEVENT_JOY1_FIRE_BUTTON;
	mice[0].buttons_eventid[1] = INPUTEVENT_JOY1_2ND_BUTTON;
	mice[0].buttons_eventid[2] = INPUTEVENT_JOY1_3RD_BUTTON;
    } else if (JSEM_ISMOUSE (1, &currprefs)) {
	mice[0].axles_eventid[0] = INPUTEVENT_MOUSE2;
        mice[0].buttons_eventid[0] = INPUTEVENT_JOY2_FIRE_BUTTON;
	mice[0].buttons_eventid[1] = INPUTEVENT_JOY2_2ND_BUTTON;
	mice[0].buttons_eventid[2] = INPUTEVENT_JOY2_3RD_BUTTON;
    }

    /* joystick compatibility code */
    joy = -1;
    if (JSEM_ISJOY0 (1, &currprefs))
	joy = 0;
    else if (JSEM_ISJOY1 (1, &currprefs))
	joy = 1;
    if (joy >= 0) {
        joysticks[joy].axles_eventid[0] = INPUTEVENT_JOY2_HORIZ;
	joysticks[joy].axles_eventid[1] = INPUTEVENT_JOY2_VERT;
	joysticks[joy].buttons_eventid[0] = INPUTEVENT_JOY2_FIRE_BUTTON;
	joysticks[joy].buttons_eventid[1] = INPUTEVENT_JOY2_2ND_BUTTON;
	joysticks[joy].buttons_eventid[2] = INPUTEVENT_JOY2_3RD_BUTTON;
    }
    if (JSEM_ISNUMPAD (1, &currprefs) || JSEM_ISCURSOR (1, &currprefs) || JSEM_ISSOMEWHEREELSE (1, &currprefs)) {
        joysticks[1].axles_eventid[0] = INPUTEVENT_JOY2_HORIZ;
	joysticks[1].axles_eventid[1] = INPUTEVENT_JOY2_VERT;
	joysticks[1].buttons_eventid[0] = INPUTEVENT_JOY2_FIRE_BUTTON;
	joysticks[1].buttons_eventid[1] = INPUTEVENT_JOY2_2ND_BUTTON;
	joysticks[1].buttons_eventid[2] = INPUTEVENT_JOY2_3RD_BUTTON;
    }

    joy = -1;
    if (JSEM_ISJOY0 (0, &currprefs))
	joy = 0;
    else if (JSEM_ISJOY1 (0, &currprefs))
	joy = 1;
    if (joy >= 0) {
        joysticks[joy].axles_eventid[0] = INPUTEVENT_JOY1_HORIZ;
	joysticks[joy].axles_eventid[1] = INPUTEVENT_JOY1_VERT;
	joysticks[joy].buttons_eventid[0] = INPUTEVENT_JOY1_FIRE_BUTTON;
	joysticks[joy].buttons_eventid[1] = INPUTEVENT_JOY1_2ND_BUTTON;
	joysticks[joy].buttons_eventid[2] = INPUTEVENT_JOY1_3RD_BUTTON;
    }
    if (JSEM_ISNUMPAD (0, &currprefs) || JSEM_ISCURSOR (0, &currprefs) || JSEM_ISSOMEWHEREELSE (0, &currprefs)) {
        joysticks[0].axles_eventid[0] = INPUTEVENT_JOY1_HORIZ;
	joysticks[0].axles_eventid[1] = INPUTEVENT_JOY1_VERT;
	joysticks[0].buttons_eventid[0] = INPUTEVENT_JOY1_FIRE_BUTTON;
	joysticks[0].buttons_eventid[1] = INPUTEVENT_JOY1_2ND_BUTTON;
	joysticks[0].buttons_eventid[2] = INPUTEVENT_JOY1_3RD_BUTTON;
    }

    /* CD32 buttons (mapped to PSX controller on USB adapter) */
    if (cd32_enabled) {
	joysticks[0].buttons_eventid[8] = INPUTEVENT_JOY2_CD32_PLAY;
	joysticks[0].buttons_eventid[7] = INPUTEVENT_JOY2_CD32_PLAY;
	joysticks[0].buttons_eventid[6] = INPUTEVENT_JOY2_CD32_PLAY;
	joysticks[0].buttons_eventid[5] = INPUTEVENT_JOY2_CD32_FFW;
	joysticks[0].buttons_eventid[4] = INPUTEVENT_JOY2_CD32_RWD;
	joysticks[0].buttons_eventid[3] = INPUTEVENT_JOY2_CD32_GREEN;
	joysticks[0].buttons_eventid[2] = INPUTEVENT_JOY2_CD32_RED;
	joysticks[0].buttons_eventid[1] = INPUTEVENT_JOY2_CD32_BLUE;
	joysticks[0].buttons_eventid[0] = INPUTEVENT_JOY2_CD32_YELLOW;
    }

    /* map PSX controller direction buttons to joystick axles */
    joysticks[0].buttons_eventid[12] = INPUTEVENT_JOY2_UP;
    joysticks[0].buttons_eventid[13] = INPUTEVENT_JOY2_RIGHT;
    joysticks[0].buttons_eventid[14] = INPUTEVENT_JOY2_DOWN;
    joysticks[0].buttons_eventid[15] = INPUTEVENT_JOY2_LEFT;

    /* joystick mouse 1 emulation (PSX controller second analog stick) */
/*
    joysticks[0].axles_eventid[2] = INPUTEVENT_MOUSE1_HORIZ;
    joysticks[0].axles_eventid[3] = INPUTEVENT_MOUSE1_VERT;
    joysticks[0].buttons_eventid[6] = INPUTEVENT_JOY1_FIRE_BUTTON;
    joysticks[0].buttons_eventid[7] = INPUTEVENT_JOY1_2ND_BUTTON;
*/
    scanjoyevents();
}

/* any ideas here?
 * is it possible to handle all native->amiga keyboard translation here?
 * native code only needs to give translation table (key code and event code)
 */

void setkeytranslation (struct native2amigakey *trans)
{
    if (!keytranstable) {
	keytranstable = trans;
	scanjoyevents();
    }
    keytranstable = trans;
}

/* keycode = -1, asciicode > 0 : only ascii code need match
 * keycode > 0, asciicode = -1: only keycode need match
 * keycode > 0, asciicode > 0 : both must match
 */
void translatekeycode (int keycode, char asciicode, int state)
{
    struct native2amigakey *na = keytranstable;
    int i = 0;
    if (!na || (keycode < 0 && asciicode < 0)) return;
    /* stupid scanning loop, any better ideas? */
    while (na->event != 0) {
	if ((keycode < 0 || na->keycode == keycode) && (asciicode < 0 || na->asciicode == asciicode)) {
	    handle_input_event (na->event, state, 1, 0);
	    return;
	}
	na++;
    }
}
