 /*
  * UAE - The Un*x Amiga Emulator
  *
  * WIN32 CDROM/HD low level access code (IOCTL)
  *
  * Copyright 2002 Toni Wilen
  *
  */
#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "uae.h"
#include "threaddep/thread.h"
#include "blkdev.h"

#include <devioctl.h>
#include <ntddcdrm.h>
#include <ntdddisk.h>

struct cd_info_win32 {
    HANDLE h;
    uae_u8 *tempbuffer;
    char drvletter;
};

static UINT errormode;

static void seterrormode (void)
{
    errormode = SetErrorMode( SEM_FAILCRITICALERRORS | SEM_NOOPENFILEERRORBOX );
}
static void reseterrormode (void)
{
    SetErrorMode (errormode);
}

static struct cd_info_win32 ciw32[MAX_TOTAL_DEVICES];

static int win32_error (const char *format,...)
{
    LPVOID lpMsgBuf;
    va_list arglist;
    char buf[1000];
    DWORD err = GetLastError();

    va_start (arglist, format );
    vsprintf (buf, format, arglist);
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
	NULL,err,MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
	(LPTSTR) &lpMsgBuf,0,NULL);
    device_debug ("%s,%d: %s ",buf, err, (char*)lpMsgBuf);
    va_end( arglist );
    return err;
}


/* pause/unpause CD audio */
static int ioctl_command_pause (int unitnum, int paused)
{
    DWORD len;
    int command = paused ? IOCTL_CDROM_PAUSE_AUDIO : IOCTL_CDROM_RESUME_AUDIO;

    seterrormode ();
    if (!DeviceIoControl(ciw32[unitnum].h, command, NULL, 0, NULL, 0, &len, NULL)) {
	reseterrormode ();
	win32_error (paused ? "IOCTL_CDROM_PAUSE_AUDIO" : "IOCTL_CDROM_RESUME_AUDIO");
	return 0;
    }
    reseterrormode ();
    return 1;
}

/* stop CD audio */
static int ioctl_command_stop (int unitnum)
{
    DWORD len;
    seterrormode ();
    if(!DeviceIoControl(ciw32[unitnum].h, IOCTL_CDROM_STOP_AUDIO, NULL, 0, NULL, 0, &len, NULL)) {
	reseterrormode ();
	win32_error ("IOCTL_CDROM_STOP_AUDIO");
	return 0;
    }
    reseterrormode ();
    return 1;
}

/* play CD audio */
static int ioctl_command_play (int unitnum, uae_u32 start, uae_u32 end, int scan)
{
    DWORD len;
    CDROM_PLAY_AUDIO_MSF pa;

    pa.StartingM = start >> 16;
    pa.StartingS = start >> 8;
    pa.StartingF = start >> 0;
    pa.EndingM = end >> 16;
    pa.EndingS = end >> 8;
    pa.EndingF = end >> 0;
    seterrormode ();
    if (!DeviceIoControl(ciw32[unitnum].h, IOCTL_CDROM_PLAY_AUDIO_MSF, &pa, sizeof(pa), NULL, 0, &len, NULL)) {
	reseterrormode ();
	win32_error ("IOCTL_CDROM_PLAY_AUDIO_MSF %02.%02.%02-%02.%02.%02",
	    pa.StartingM, pa.StartingS, pa.StartingF, pa.EndingM, pa.EndingS, pa.EndingF );
	return 0;
    }
    reseterrormode ();
    return 1;
}

/* read qcode */
static uae_u8 *ioctl_command_qcode (int unitnum)
{
    SUB_Q_CHANNEL_DATA qcd;
    DWORD len;
    ULONG in = 1;
    uae_u8 *p = ciw32[unitnum].tempbuffer;
    
    memset (p, 0, 4 + 12);
    p[1] = 0x15;
    p[3] = 12;
    reseterrormode ();
    if(!DeviceIoControl(ciw32[unitnum].h, IOCTL_CDROM_READ_Q_CHANNEL, &in, sizeof(in), &qcd, sizeof (qcd), &len, NULL)) {
	reseterrormode ();
	win32_error ("IOCTL_CDROM_READ_Q_CHANNEL");
	return 0;
    }
    p[1] = qcd.CurrentPosition.Header.AudioStatus;
    p += 4;
    p[1] = (qcd.CurrentPosition.Control << 0) | (qcd.CurrentPosition.ADR << 4);
    p[2] = qcd.CurrentPosition.TrackNumber;
    p[3] = qcd.CurrentPosition.IndexNumber;
    p[5] = qcd.CurrentPosition.AbsoluteAddress[1];
    p[6] = qcd.CurrentPosition.AbsoluteAddress[2];
    p[7] = qcd.CurrentPosition.AbsoluteAddress[3];
    p[9] = qcd.CurrentPosition.TrackRelativeAddress[1];
    p[10] = qcd.CurrentPosition.TrackRelativeAddress[2];
    p[11] = qcd.CurrentPosition.TrackRelativeAddress[3];
    reseterrormode ();
    return ciw32[unitnum].tempbuffer;
}

static uae_u8 *ioctl_command_read (int unitnum, int sector)
{
    DWORD dtotal;
    seterrormode ();
    if (SetFilePointer (ciw32[unitnum].h, sector  * 2048, 0, FILE_BEGIN) == INVALID_SET_FILE_POINTER) {
	reseterrormode ();
	win32_error ("SetFilePointer");
	return 0;
    }
    if (!ReadFile (ciw32[unitnum].h, ciw32[unitnum].tempbuffer, 2048, &dtotal, 0)) {
	reseterrormode ();
	win32_error ("ReadFile");
	return 0;
    }
    reseterrormode ();
    return ciw32[unitnum].tempbuffer;
}

/* read toc and geometry */
static uae_u8 *ioctl_command_toc (int unitnum)
{
    CDROM_TOC toc;
    DWORD len;
    int i;
    uae_u8 *p = ciw32[unitnum].tempbuffer;

    seterrormode ();
    if (!DeviceIoControl(ciw32[unitnum].h, IOCTL_CDROM_READ_TOC, NULL, 0, &toc, sizeof(toc), &len, NULL)) {
	reseterrormode ();
	return 0;
    }
    p[0] = ((toc.LastTrack + 4) * 11) >> 8;
    p[1] = ((toc.LastTrack + 4) * 11) & 0xff;
    p[2] = 1;
    p[3] = toc.LastTrack;
    p += 4;
    memset (p, 0, 11);
    p[0] = 1;
    p[1] = (toc.TrackData[0].Control << 0) | (toc.TrackData[0].Adr << 4);
    p[3] = 0xa0;
    p[8] = 1;
    p += 11;
    memset (p, 0, 11);
    p[0] = 1;
    p[1] = 0x10;
    p[3] = 0xa1;
    p[8] = toc.LastTrack;
    p += 11;
    memset (p, 0, 11);
    p[0] = 1;
    p[1] = 0x10;
    p[3] = 0xa2;
    p[8] = toc.TrackData[toc.LastTrack].Address[1];
    p[9] = toc.TrackData[toc.LastTrack].Address[2];
    p[10] = toc.TrackData[toc.LastTrack].Address[3];
    p += 11;
    for (i = 0; i < toc.LastTrack; i++) {
	memset (p, 0, 11);
	p[0] = 1;
	p[1] = (toc.TrackData[i].Control << 0) | (toc.TrackData[i].Adr << 4);
	p[2] = 0;
	p[3] = i + 1;
	p[8] = toc.TrackData[i].Address[1];
	p[9] = toc.TrackData[i].Address[2];
	p[10] = toc.TrackData[i].Address[3];
	p += 11;
    }
    reseterrormode ();
    return ciw32[unitnum].tempbuffer;
}

/* open device level access to cd rom drive */
static int sys_cddev_open (int unitnum)
{
    DWORD len;
    char dev[10];
    struct cd_info_win32 *ciw = &ciw32[unitnum];
    DWORD flags;

    sprintf (dev,"\\\\.\\%c:", ciw->drvletter);
    /* buffer must be page aligned for device access */
    ciw->tempbuffer = VirtualAlloc (NULL, 4096, MEM_COMMIT, PAGE_READWRITE);
    if (!ciw->tempbuffer) {
	device_debug ("failed to allocate buffer");
	return 1;
    }
    flags = GENERIC_READ;
    ciw->h = CreateFile(dev, flags, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (ciw->h == INVALID_HANDLE_VALUE) {
	flags |= GENERIC_WRITE;
        ciw->h = CreateFile(dev, flags, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if (ciw->h == INVALID_HANDLE_VALUE) {
	    device_debug ("failed to open cd-rom handle (%s)\n",dev);
	    goto error;
	}
    }
    DeviceIoControl(ciw->h, IOCTL_STORAGE_CHECK_VERIFY, NULL, 0, NULL, 0, &len, NULL);
    device_debug ("device '%s' opened succesfully (unit number=%d)\n",dev,unitnum);
    sys_command_stop (unitnum);
    return 0;
error:
    win32_error ("CreateFile");
    VirtualFree (ciw->tempbuffer, 0, MEM_RELEASE);
    ciw->tempbuffer = NULL;
    CloseHandle (ciw->h);
    ciw->h = NULL;
    return -1;
}

/* close device handle */
void sys_cddev_close (int unitnum)
{
    CloseHandle (ciw32[unitnum].h);
    ciw32[unitnum].h = NULL;
    VirtualFree (ciw32[unitnum].tempbuffer, 0, MEM_RELEASE);
    ciw32[unitnum].tempbuffer = NULL;
}

static int unitcheck (int unitnum)
{
    if (unitnum >= MAX_TOTAL_DEVICES)
	return 0;
    if (ciw32[unitnum].drvletter == 0)
	return 0;
    return 1;
}

static int open_device (int unitnum)
{
    if (!unitcheck (unitnum))
	return 0;
    if (sys_cddev_open (unitnum) == 0)
	return 1;
    return 0;
}
static void close_device (int unitnum)
{
    sys_cddev_close (unitnum);
}

static void close_bus (void)
{
}

static int total_devices;

static int open_bus (int flags)
{
    int dwDriveMask;
    int drive, i;
    char tmp[10];

    for (i = 0; i < MAX_TOTAL_DEVICES; i++)
	ciw32[i].drvletter = 0;
    total_devices = 0;
    dwDriveMask = GetLogicalDrives();
    dwDriveMask >>= 2; // Skip A and B drives...
    for( drive = 'C'; drive <= 'Z'; drive++) {
	if (dwDriveMask & 1) {
	    sprintf( tmp, "%c:\\", drive );
	    if (flags == INQ_ROMD && GetDriveType (tmp) == DRIVE_CDROM) {
		ciw32[total_devices++].drvletter = drive;
	    }
	}
	dwDriveMask >>= 1;
    }
    return total_devices;
}

static int media_device (int unitnum)
{
    int ret = 0;

    if (!unitcheck (unitnum))
	return 0;
    if (ciw32[unitnum].h == INVALID_HANDLE_VALUE)
	return ret;
    ret = ioctl_command_toc (unitnum) == 0 ? 0 : 1;
    if (GetLastError() == 34) {
	write_log ("media change, re-opening device\n");
	close_device (unitnum);
	open_device (unitnum);
    }
    return ret;
}

static struct device_info *info_device (int unitnum, struct device_info *di)
{
    if (!unitcheck (unitnum))
	return 0;
    di->bus = unitnum;
    di->target = 0;
    di->lun = 0;
    di->media_inserted = media_device (unitnum);
    di->write_protected = 1;
    di->sectorsize = 2048;
    di->type = INQ_ROMD;
    return di;
}

struct device_functions devicefunc_win32_ioctl = {
    open_bus, close_bus, open_device, close_device, info_device,
    0, 0, 0,
    ioctl_command_pause, ioctl_command_stop, ioctl_command_play, ioctl_command_qcode,
    ioctl_command_toc, ioctl_command_read
};

