 /*
  * UAE - The Un*x Amiga Emulator
  *
  * scsi.device emulation
  *
  * Copyright 1995 Bernd Schmidt
  * Copyright 1999 Patrick Ohly
  * Copyright 2001 Brian King
  * Copyright 2002 Toni Wilen
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "uae.h"
#include "threaddep/thread.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "events.h"
#include "newcpu.h"
#include "autoconf.h"
#include "execlib.h"
#include "native2amiga.h"
#include "blkdev.h"

#define dev_debug
#undef CDDEV_COMMANDS

#define MAX_ASYNC_REQUESTS 20

#define CMD_INVALID	0
#define CMD_RESET	1
#define CMD_READ	2
#define CMD_WRITE	3
#define CMD_UPDATE	4
#define CMD_CLEAR	5
#define CMD_STOP	6
#define CMD_START	7
#define CMD_FLUSH	8

#define ASYNC_REQUEST_NONE 0
#define ASYNC_REQUEST_TEMP 1
#define ASYNC_REQUEST_CHANGEINT 10
#define ASYNC_REQUEST_FRAMEINT 11
#define ASYNC_REQUEST_PLAY 12
#define ASYNC_REQUEST_READXL 13
#define ASYNC_REQUEST_FRAMECALL 14

struct devstruct {
    int unitnum, aunit;
    int bus, target, lun;
    char label[64];
    int opencnt;
    int changenum;
    int allow_scsi_direct;
    int drivetype;
    int numtracks;
    int sectorsize;
    volatile uaecptr d_request[MAX_ASYNC_REQUESTS];
    volatile int d_request_type[MAX_ASYNC_REQUESTS];
    volatile uae_u32 d_request_data[MAX_ASYNC_REQUESTS];

    smp_comm_pipe requests;
    uae_thread_id tid;
    int thread_running;
    uae_sem_t sync_sem;
};

static struct devstruct devst[MAX_TOTAL_DEVICES];

static struct device_info *devinfo (int unitnum, struct device_info *di)
{
    return device_func->info (unitnum, di);
}

static void io_log (char *msg, uaecptr request)
{
#if 0
    write_log ("%s: %08X %d %08.8X %d %d\n", msg, request,get_word(request + 28),get_long(request + 40),get_long(request + 36),get_long(request + 44) );
#endif
}

void memcpyha (uae_u32 dst, char *src, int size)
{
    while (size--)
	put_byte (dst++, *src++);
}

static struct devstruct *getdevstruct (int unit)
{
    int i;
    for (i = 0; i < MAX_TOTAL_DEVICES; i++) {
	if (unit >= 0 && devst[i].aunit == unit) return &devst[i];
    }
    return 0;
}

static void *dev_thread (void *devs);
static int start_thread (struct devstruct *dev)
{
    if (dev->thread_running)
        return 1;
    init_comm_pipe (&dev->requests, 10, 1);
    uae_sem_init (&dev->sync_sem, 0, 0);
    uae_start_thread (dev_thread, dev, &dev->tid);
    uae_sem_wait (&dev->sync_sem);
    return dev->thread_running;
}


static void dev_close_2 (struct devstruct *dev)
{
    if (!dev->opencnt) return;
    dev->opencnt--;
    if (!dev->opencnt) {
	device_func->closedev (dev->unitnum);
        write_comm_pipe_u32 (&dev->requests, 0, 1);
    }
}

static uae_u32 dev_close (void)
{
    uae_u32 request = m68k_areg (regs, 1);
    int unit = get_long (request + 24);
    struct devstruct *dev = getdevstruct (unit);

    dev_debug ("dev close, req=%08.8X, unit=%d\n",request, unit);
    if (!dev)
	return 0;
    dev_close_2 (dev);
    put_word (m68k_areg(regs, 6) + 32, get_word (m68k_areg(regs, 6) + 32) - 1);
    return 0;
}

static int openfail (uaecptr ioreq, int error)
{
    put_long (ioreq + 20, -1);
    put_byte (ioreq + 31, error);
    return (uae_u32)-1;
}

static uae_u32 dev_open_2 (void)
{
    uaecptr ioreq = m68k_areg(regs, 1);
    uae_u32 unit = m68k_dreg (regs, 0);
    struct devstruct *dev = getdevstruct (unit);

    dev_debug ("opening %s.device, ioreq=%08.8X unit=%d\n", "uaescsi", ioreq, unit);
    if (!dev)
	return openfail (ioreq, 32); /* badunitnum */
    if (!dev->opencnt) {
	if (!device_func->opendev (dev->unitnum))
	    return openfail (ioreq, 32);
	start_thread (dev);
    }
    dev->opencnt++;

    dev_debug ("device opened succesfully\n");
    put_word (m68k_areg(regs, 6)+32, get_word (m68k_areg(regs, 6)+32) + 1);
    put_long (ioreq + 24, unit);
    put_byte (ioreq + 31, 0);
    put_byte (ioreq + 8, 7);
    return 0;
}

static uae_u32 dev_open (void)
{
    return dev_open_2 ();
}

static uae_u32 dev_expunge (void)
{
    return 0;
}

static int is_async_request (struct devstruct *dev, uaecptr request)
{
    int i;
    i = 0;
    while (i < MAX_ASYNC_REQUESTS) {
	if (dev->d_request[i] == request) return 1;
	i++;
    }
    return 0;
}

static int add_async_request (struct devstruct *dev, uaecptr request, int type, uae_u32 data)
{
    int i;
    i = 0;
    while (i < MAX_ASYNC_REQUESTS) {
	if (dev->d_request[i] == request) {
	    dev->d_request_type[i] = type;
	    dev->d_request_data[i] = data;
	    return 0;
	}
	i++;
    }
    i = 0;
    while (i < MAX_ASYNC_REQUESTS) {
	if (dev->d_request[i] == 0) {
	    dev->d_request[i] = request;
	    dev->d_request_type[i] = type;
	    dev->d_request_data[i] = data;
	    return 0;
	}
	i++;
    }
    return -1;
}

static int release_async_request (struct devstruct *dev, uaecptr request)
{
    int i = 0;
    while (i < MAX_ASYNC_REQUESTS) {
	if (dev->d_request[i] == request) {
	    int type = dev->d_request_type[i];
	    dev->d_request[i] = 0;
	    dev->d_request_data[i] = 0;
	    dev->d_request_type[i] = 0;
	    return type;
	}
	i++;
    }
    return -1;
}

static void abort_async (struct devstruct *dev, uaecptr request, int errcode, int type)
{
    int i;
    i = 0;
    while (i < MAX_ASYNC_REQUESTS) {
	if (dev->d_request[i] == request && dev->d_request_type[i] == ASYNC_REQUEST_TEMP) {
	    /* ASYNC_REQUEST_TEMP = request is processing */
	    Sleep (10);
	    i = 0;
	    continue;
	}
	i++;
    }
    i = release_async_request (dev, request);
    if (i >= 0)
	dev_debug("asyncronous request=%08.8X aborted, error=%d\n", request, errcode);
}

static int command_read (struct devstruct *dev, uaecptr data, int offset, int length, uae_u32 *io_actual)
{
    uae_u8 *temp;
    int len, sector;

    int startoffset = offset % dev->sectorsize;
    offset -= startoffset;
    sector = offset / dev->sectorsize;
    *io_actual = 0;
    while (length > 0) {
	temp = sys_command_read (dev->unitnum, sector);
	if (!temp) return 20;
	if (startoffset > 0) {
	    len = dev->sectorsize - startoffset;
	    if (len > length) len = length;
	    memcpyha (data, temp + startoffset, len);
	    length -= len;
	    data += len;
	    startoffset = 0;
	    *io_actual += len;
	} else if (length >= dev->sectorsize) {
	    len = dev->sectorsize;
	    memcpyha (data, temp, len);
	    length -= len;
	    data += len;
	    *io_actual += len;
	} else {
	    memcpyha (data, temp, length);
	    *io_actual += length;
	    length = 0;
	}
	sector++;
    }
    return 0;
}


static int dev_do_io (struct devstruct *dev, uaecptr request)
{
    uae_u32 command;
    uae_u32 io_data = get_long (request + 40); // 0x28
    uae_u32 io_length = get_long (request + 36); // 0x24
    uae_u32 io_actual = get_long (request + 32); // 0x20
    uae_u32 io_offset = get_long (request + 44); // 0x2c
    uae_u32 io_error = 0;
    int async = 0;
    struct device_info di;

    io_log ("dev_io",request);
    command = get_word (request+28);
    switch (command)
    {
#ifdef CDDEV_COMMANDS
	case CMD_READ:
	io_error = command_read (dev, io_data, io_offset, io_length, &io_actual);
	break;
	case CMD_WRITE:
	io_error = 28; /* writeprotect */
	break;
	case CMD_UPDATE:
	case CMD_CLEAR:
	case CMD_FLUSH:
	case 9: /* Motor */
	case 10: /* Seek */
	io_actual = 0;
	break;
	case 11: /* Format */
	io_error = 28; /* writeprotect */
	break;
	case 12: /* Remove */
	io_actual = 0;
	break;
	case 13: /* ChangeNum */
	io_actual = dev->changenum;
	break;
	case 14: /* ChangeStatus */
	io_actual = devinfo(dev->unitnum, &di)->media_inserted ? -1 : 0;
	break;
	case 15: /* ProtStatus */
	io_actual = devinfo(dev->unitnum, &di)->write_protected ? -1 : 0;
	break;
	case 18: /* GetDriveType */
	io_actual = dev->drivetype;
	break;
	case 19: /* GetNumTracks */
	io_actual = dev->numtracks;
	break;
	case 20: /* AddChangeInt */
	io_error = add_async_request (dev, request, ASYNC_REQUEST_CHANGEINT, io_data);
	if (!io_error)
	    async = 1;
	break;
	case 21: /* RemChangeInt */
	release_async_request (dev, request);
	break;
#endif
	case 28: /* HD_SCSICMD */
	if (dev->allow_scsi_direct && device_func->exec_direct) {
	    uae_u32 sdd = get_long (request + 40);
	    io_error = device_func->exec_direct (dev->unitnum, sdd);
	    dev_debug ("scsidev: did io: sdd     = 0x%x\n", sdd);
	    dev_debug ("scsidev: did io: request = %08lx\n", (unsigned long)request);
	    dev_debug ("scsidev: did io: error   = %d\n", (int)get_word(request+31));
	} else
	    io_error = -3;
        break;
	default:
	io_error = -3;
	break;
    }
    put_long (request + 32, io_actual);
    put_byte (request + 31, io_error);
    return async;
}


static int dev_can_quick (uae_u32 command)
{
    switch (command)
    {
	case CMD_RESET:
	case CMD_STOP:
	case CMD_START:
	case 14:
	case 15:
	case 18:
	case 19:
	return 1;
    }
    return 0;
}

static int dev_canquick (struct devstruct *dev, uaecptr request)
{
    uae_u32 command = get_word (request + 28);
    return dev_can_quick (command);
}

static uae_u32 dev_beginio (void)
{
    uae_u32 request = m68k_areg(regs, 1);
    int unit = get_long (request + 24);
    uae_u8 flags = get_byte (request + 30);
    int command = get_word (request + 28);
    struct devstruct *dev = getdevstruct (unit);

    put_byte (request+8, NT_MESSAGE);
    if (!dev) {
	put_byte (request + 31, 32);
	return get_byte (request + 31);
    }
    put_byte (request+31, 0);
    if ((flags & 1) && dev_canquick (dev, request)) {
	if (dev_do_io (dev, request))
	    dev_debug ("command %d bug with IO_QUICK\n", command);
	return get_byte (request + 31);
    } else {
        add_async_request (dev, request, ASYNC_REQUEST_TEMP, 0);
        put_byte (request+30, get_byte (request + 30) & ~1);
	write_comm_pipe_u32 (&dev->requests, request, 1);
	return 0;
    }
}

static void *dev_thread (void *devs)
{
    struct devstruct *dev = devs;

    dev->thread_running = 1;
    uae_sem_post (&dev->sync_sem);
    for (;;) {
	uaecptr request;
	request = (uaecptr)read_comm_pipe_u32_blocking (&dev->requests);
        if (!request) {
	    dev->thread_running = 0;
	    uae_sem_post (&dev->sync_sem);
	    return 0;
	} else if (dev_do_io (dev, request) == 0) {
            put_byte (request+30, get_byte (request + 30) & ~1);
            uae_ReplyMsg (request);
	    release_async_request (dev, request);
	} else {
	    dev_debug("async request %08.8X\n", request);
	}
    }
    return 0;
}

static uae_u32 dev_init (void)
{
    uae_u32 base = m68k_dreg (regs,0);
    dev_debug("cd.device init\n");
    return base;
}

static uae_u32 dev_abortio (void)
{
    uae_u32 request = m68k_areg(regs, 1);
    int unit = get_long (request + 24);
    struct devstruct *dev = getdevstruct (unit);

    if (!dev) {
	put_byte (request + 31, 32);
	return get_byte (request + 31);
    }
    put_byte (request + 31, -2);
    dev_debug ("abortio unit=%d, request=%08.8X\n", unit, request);
    abort_async (dev, request, -2, 0);
    return 0;
}

#define BTL2UNIT(bus, target, lun) (2 * (bus) + (target) / 8) * 100 + (lun) * 10 + (target % 8)

static void dev_reset (void)
{
    int i, j;
    struct devstruct *dev;
    struct device_info di1, *di2;
    int fixunits = 1;

    if (!currprefs.scsi)
	return;
    device_func_init (DEVICE_TYPE_SCSI);
    for (i = 0; i < MAX_TOTAL_DEVICES; i++) {
	dev = &devst[i];
	if (dev->opencnt > 0) {
	    for (j = 0; j < MAX_ASYNC_REQUESTS; j++) {
	        uaecptr request;
		if (request = dev->d_request[i]) 
		    abort_async (dev, request, 0, 0);
	    }
	    dev->opencnt = 1;
	    dev_close_2 (dev);
	}
	dev->unitnum = dev->aunit = -1;
	di2 = device_func->info (i, &di1);
	if (di2) {
	    dev->allow_scsi_direct = 1;
	    dev->unitnum = i;
	    dev->bus = di2->bus;
	    dev->target = di2->target;
	    dev->lun = di2->lun;
	    dev->aunit = BTL2UNIT (di2->bus, di2->target, di2->lun);
	    strcpy (dev->label, di2->label);
	    if (dev->target != 0 || dev->lun != 0)
		fixunits = 0;
	    dev->drivetype = di2->type;
	    if (di2->type == INQ_ROMD) {
		dev->numtracks = 1;
		dev->changenum = 0;
		dev->sectorsize = di2->sectorsize;
	    }
	}
    }
    if (fixunits) {
        for (i = 0; i < MAX_TOTAL_DEVICES; i++) {
	    dev = &devst[i];
	    if (dev->unitnum >= 0)
		dev->aunit = BTL2UNIT (0, dev->bus, 0);
	}
    }
    for (i = 0; i < MAX_TOTAL_DEVICES; i++) {
        dev = &devst[i];
        if (dev->unitnum >= 0)
	    write_log("%s = uaescsi.device:%d\n", dev->label, dev->aunit);
    }
}

static uaecptr ROM_scsidev_resname = 0,
    ROM_scsidev_resid = 0,
    ROM_scsidev_init = 0;

uaecptr scsidev_startup (uaecptr resaddr)
{
    if (!currprefs.scsi)
	return resaddr;
    dev_debug ("scsidev_startup(0x%x)\n", resaddr);
    /* Build a struct Resident. This will set up and initialize
     * the uaescsi.device */
    put_word(resaddr + 0x0, 0x4AFC);
    put_long(resaddr + 0x2, resaddr);
    put_long(resaddr + 0x6, resaddr + 0x1A); /* Continue scan here */
    put_word(resaddr + 0xA, 0x8101); /* RTF_AUTOINIT|RTF_COLDSTART; Version 1 */
    put_word(resaddr + 0xC, 0x0305); /* NT_DEVICE; pri 05 */
    put_long(resaddr + 0xE, ROM_scsidev_resname);
    put_long(resaddr + 0x12, ROM_scsidev_resid);
    put_long(resaddr + 0x16, ROM_scsidev_init); /* calls scsidev_init */
    resaddr += 0x1A;

    return resaddr;
}

void scsidev_install (void)
{
    uae_u32 functable, datatable;
    uae_u32 initcode, openfunc, closefunc, expungefunc;
    uae_u32 beginiofunc, abortiofunc;

    if (!currprefs.scsi)
	return;
    dev_debug ("scsidev_install(): 0x%x\n", here ());

    ROM_scsidev_resname = ds ("uaescsi.device");
    ROM_scsidev_resid = ds ("UAE scsi.device 0.2");

    /* initcode */
    initcode = here ();
    calltrap (deftrap (dev_init)); dw (RTS);
    
    /* Open */
    openfunc = here ();
    calltrap (deftrap (dev_open)); dw (RTS);

    /* Close */
    closefunc = here ();
    calltrap (deftrap (dev_close)); dw (RTS);

    /* Expunge */
    expungefunc = here ();
    calltrap (deftrap (dev_expunge)); dw (RTS);

    /* BeginIO */
    beginiofunc = here ();
    calltrap (deftrap (dev_beginio));
    dw (RTS);

    /* AbortIO */
    abortiofunc = here ();
    calltrap (deftrap (dev_abortio)); dw (RTS);

    /* FuncTable */
    functable = here ();
    dl (openfunc); /* Open */
    dl (closefunc); /* Close */
    dl (expungefunc); /* Expunge */
    dl (EXPANSION_nullfunc); /* Null */
    dl (beginiofunc); /* BeginIO */
    dl (abortiofunc); /* AbortIO */
    dl (0xFFFFFFFFul); /* end of table */

    /* DataTable */
    datatable = here ();
    dw (0xE000); /* INITBYTE */
    dw (0x0008); /* LN_TYPE */
    dw (0x0300); /* NT_DEVICE */
    dw (0xC000); /* INITLONG */
    dw (0x000A); /* LN_NAME */
    dl (ROM_scsidev_resname);
    dw (0xE000); /* INITBYTE */
    dw (0x000E); /* LIB_FLAGS */
    dw (0x0600); /* LIBF_SUMUSED | LIBF_CHANGED */
    dw (0xD000); /* INITWORD */
    dw (0x0014); /* LIB_VERSION */
    dw (0x0004); /* 0.4 */
    dw (0xD000); /* INITWORD */
    dw (0x0016); /* LIB_REVISION */
    dw (0x0000); /* end of table already ??? */
    dw (0xC000); /* INITLONG */
    dw (0x0018); /* LIB_IDSTRING */
    dl (ROM_scsidev_resid);
    dw (0x0000); /* end of table */

    ROM_scsidev_init = here ();
    dl (0x00000100); /* size of device base */
    dl (functable);
    dl (datatable);
    dl (initcode);
}

void scsidev_start_threads (void)
{
    if (!currprefs.scsi) /* quite useless.. */
	return;
    dev_debug ("scsidev_start_threads()\n");
}

void scsidev_reset (void)
{
    dev_reset ();
}